from torchvision.datasets import *
from torchvision.datasets.vision import StandardTransform

from .oxford_iiit_pet_plus import OxfordIIITPetPlus
from .stack import StackDataset
from .stanford_dogs import StanfordDogs
